import base64
import gzip
import tempfile
from contextlib import contextmanager
from datetime import datetime
from functools import cached_property
from io import BytesIO
from os import PathLike
from pathlib import Path
from typing import Dict, Optional, Tuple, Union

import magic
import pypdfium2 as pdfium
from PIL import Image
from pydantic import (
    BaseModel,
    Field,
    PositiveInt,
    PrivateAttr,
    computed_field,
    field_serializer,
    field_validator,
)

from docprompt._exec.ghostscript import (
    compress_pdf_to_bytes,
)

DEFAULT_DPI = 100


def get_page_render_size_from_bytes(
    file_bytes: bytes, page_number: int, dpi: int = DEFAULT_DPI
):
    """
    Returns the render size of a page in pixels
    """

    pdf = pdfium.PdfDocument(BytesIO(file_bytes))
    page = pdf.get_page(page_number)

    mediabox = page.get_mediabox()

    base_width = int(mediabox[2] - mediabox[0])
    base_height = int(mediabox[3] - mediabox[1])

    width = int(base_width * dpi / 72)
    height = int(base_height * dpi / 72)

    return width, height


class PdfDocument(BaseModel):
    """
    Represents a PDF document
    """

    name: str = Field(description="The name of the document")
    file_bytes: bytes = Field(description="The bytes of the document", repr=False)
    file_path: Optional[str] = None

    _raster_cache: Dict[int, Dict[int, bytes]] = PrivateAttr(default_factory=dict)

    def __len__(self):
        return self.num_pages

    def __hash__(self):
        return hash(self.document_hash)

    @computed_field
    @cached_property
    def page_count(self) -> PositiveInt:
        from docprompt.utils.util import get_page_count

        return get_page_count(self.file_bytes)

    @property
    def num_pages(self):
        return self.page_count

    @computed_field
    @cached_property
    def document_hash(self) -> str:
        from docprompt.utils.util import hash_from_bytes

        return hash_from_bytes(self.file_bytes)

    @field_serializer("file_bytes")
    def serialize_file_bytes(self, v: bytes, _info):
        compressed = gzip.compress(v)

        return base64.b64encode(compressed).decode("utf-8")

    @field_validator("file_bytes")
    def validate_file_bytes(cls, v: bytes):
        if not isinstance(v, bytes):
            raise ValueError("File bytes must be bytes")

        if len(v) == 0:
            raise ValueError("File bytes must not be empty")

        if magic.from_buffer(v, mime=True) == "text/plain":
            v = base64.b64decode(v, validate=True)

        if magic.from_buffer(v, mime=True) == "application/gzip":
            v = gzip.decompress(v)

        if magic.from_buffer(v, mime=True) != "application/pdf":
            raise ValueError("File bytes must be a PDF")

        return v

    @classmethod
    def from_path(cls, file_path: Union[PathLike, str]):
        file_path = Path(file_path)

        if not file_path.is_file():
            raise ValueError(f"File path {file_path} is not a file")

        file_bytes = file_path.read_bytes()

        return cls(name=file_path.name, file_path=str(file_path), file_bytes=file_bytes)

    @classmethod
    def from_bytes(cls, file_bytes: bytes, name: Optional[str] = None):
        if name is None:
            name = f"PDF-{datetime.now().isoformat()}.pdf"

        return cls(name=name, file_bytes=file_bytes)

    def get_bytes(self) -> bytes:
        return self.file_bytes  # Deprecated

    @property
    def path(self):
        return self.file_path

    def get_page_render_size(
        self, page_number: int, dpi: int = DEFAULT_DPI
    ) -> Tuple[int, int]:
        """
        Returns the render size of a page in pixels
        """
        return get_page_render_size_from_bytes(self.get_bytes(), page_number, dpi=dpi)

    def to_compressed_bytes(self, compression_kwargs: dict = {}) -> bytes:
        """
        Compresses the document using Ghostscript
        """
        with self.as_tempfile() as temp_path:
            return compress_pdf_to_bytes(temp_path, **compression_kwargs)

    def rasterize_page(
        self,
        page_number: int,
        *,
        dpi: int = DEFAULT_DPI,
        downscale_size: Optional[Tuple[int, int]] = None,
        use_cache: bool = True,
    ) -> bytes:
        """
        Rasterizes a page of the document using Ghostscript
        """
        generated_image = False

        if page_number < 0 or page_number > self.num_pages:
            raise ValueError(f"Page number must be between 0 and {self.num_pages}")

        if use_cache and self._raster_cache.get(dpi, {}).get(page_number):
            rastered = self._raster_cache[dpi][page_number]
        else:
            pdf = pdfium.PdfDocument(BytesIO(self.file_bytes))
            page = pdf[page_number - 1]

            bitmap = page.render(scale=(1 / 72) * dpi)

            pil = bitmap.to_pil().convert("RGB")

            img_bytes = BytesIO()
            pil.save(img_bytes, format="PNG")
            rastered = img_bytes.getvalue()

            generated_image = True

        if use_cache and generated_image:
            self._raster_cache.setdefault(dpi, {})
            self._raster_cache[dpi][page_number] = rastered

        if downscale_size:
            with Image.open(BytesIO(rastered)) as img:
                width, height = img.size
                if not (width < downscale_size[0] or height < downscale_size[1]):
                    img = img.resize(downscale_size)
                    img_bytes = BytesIO()
                    img.save(img_bytes, format="PNG")
                    rastered = img_bytes.getvalue()

        return rastered

    def rasterize_page_to_data_uri(
        self,
        page_number: int,
        *,
        dpi: int = DEFAULT_DPI,
        downscale_size: Optional[Tuple[int, int]] = None,
        use_cache: bool = True,
    ) -> str:
        """
        Rasterizes a page of the document using Ghostscript and returns a data URI, which can
        be embedded into HTML or passed to large language models
        """
        image_bytes = self.rasterize_page(
            page_number,
            dpi=dpi,
            downscale_size=downscale_size,
            use_cache=use_cache,
        )
        return f"data:image/png;base64,{base64.b64encode(image_bytes).decode('utf-8')}"

    def rasterize_pdf(
        self,
        dpi: int = DEFAULT_DPI,
        use_cache: bool = True,
    ) -> Dict[int, bytes]:
        """
        Rasterizes the entire document using Ghostscript
        """
        if (
            use_cache
            and self._raster_cache.get(dpi)
            and len(self._raster_cache[dpi]) == self.num_pages
        ):
            return self._raster_cache[dpi]

        result = {}

        for page_number in range(1, self.num_pages + 1):
            result[page_number] = self.rasterize_page(
                page_number, dpi=dpi, use_cache=False
            )

        if use_cache:
            self._raster_cache[dpi] = result.copy()  # Shallow copy should be OK

        return result

    def split(self, start: Optional[int] = None, stop: Optional[int] = None):
        """
        Splits a document into multiple documents
        """
        if start is None and stop is None:
            raise ValueError("Must specify either start or stop")

        start = start or 0

        from docprompt.utils.splitter import split_pdf_to_bytes

        split_bytes = split_pdf_to_bytes(
            self.file_bytes, start_page=start, stop_page=stop
        )

        return Document.from_bytes(split_bytes, name=self.name)

    def as_tempfile(self, **kwargs):
        """
        Returns a tempfile of the document
        """

        @contextmanager
        def tempfile_context() -> str:
            tempfile_kwargs = {"mode": "wb", "delete": True, "suffix": ".pdf", **kwargs}

            with tempfile.NamedTemporaryFile(**tempfile_kwargs) as f:
                f.write(self.file_bytes)
                f.flush()
                yield f.name

        return tempfile_context()

    def write_to_path(self, path: Union[PathLike, str], **kwargs):
        """
        Writes the document to a path
        """
        path = Path(path)

        if path.is_dir():
            path = path / self.name

        with path.open("wb") as f:
            f.write(self.file_bytes)


Document = PdfDocument  # Alias



from .util import PdfFixture


PDF_FIXTURES = [
    PdfFixture(
        name="1.pdf",
        page_count=6,
        file_hash="121ffed4336e6129e97ee3c4cb747864",
        ocr_name="1_ocr.json",
    ),
]



"""
The core primatives for any language model interfacing. Docprompt uses these for the prompt garden, but
supports free conversion to and from these types from other libaries.
"""

from typing import Literal, Union, Optional
from pydantic import BaseModel, model_validator


class OpenAIImageURL(BaseModel):
    url: str


class OpenAIComplexContent(BaseModel):
    type: Literal["text", "image_url"]
    text: Optional[str] = None
    image_url: Optional[OpenAIImageURL] = None

    @model_validator(mode="after")
    def validate_content(cls, v):
        if v.type == "text" and v.text is None:
            raise ValueError("Text content must be provided when type is 'text'")
        if v.type == "image_url" and v.image_url is None:
            raise ValueError(
                "Image URL content must be provided when type is 'image_url'"
            )

        if v.text is not None and v.image_url is not None:
            raise ValueError("Only one of text or image_url can be provided")

        return v


class OpenAIMessage(BaseModel):
    role: Literal["system", "user", "assistant"]
    content: Union[str, list[OpenAIComplexContent]]

    def to_langchain_message(self):
        try:
            from langchain.schema import SystemMessage, HumanMessage, AIMessage
        except ImportError:
            raise ImportError(
                "Could not import langchain.schema. Install with `docprompt[langchain]`"
            )

        role_mapping = {
            "system": SystemMessage,
            "user": HumanMessage,
            "assistant": AIMessage,
        }

        dumped = self.model_dump(mode="json", exclude_unset=True, exclude_none=True)

        return role_mapping[self.role](content=dumped["content"])

    def to_openai(self):
        return self.model_dump(mode="json", exclude_unset=True, exclude_none=True)

    def to_llamaindex_chat_message(self):
        try:
            from llama_index.core.base.llms.types import ChatMessage, MessageRole
        except ImportError:
            raise ImportError(
                "Could not import llama_index.core. Install with `docprompt[llamaindex]`"
            )

        role_mapping = {
            "system": MessageRole.SYSTEM,
            "user": MessageRole.USER,
            "assistant": MessageRole.ASSISTANT,
        }

        dumped = self.model_dump(mode="json", exclude_unset=True, exclude_none=True)

        return ChatMessage.from_str(
            content=dumped["content"], role=role_mapping[self.role]
        )



from typing import Optional, TYPE_CHECKING, Union, Literal, Dict, List
from docprompt.schema.document import Document

from docprompt.tasks.base import AbstractTaskProvider
from docprompt.utils.splitter import pdf_split_iter_with_max_bytes
from ..base import CAPABILITIES
from .result import OcrPageResult
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
import multiprocessing
from threading import Lock

from docprompt.schema.layout import (
    DirectionChoices,
    TextBlock,
    NormBBox,
    TextSpan,
    BoundingPoly,
    Point,
    SegmentLevels,
    TextBlockMetadata,
)
from tenacity import retry, stop_after_attempt, wait_exponential
import tqdm
import logging

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from google.cloud import documentai

    from docprompt.schema.document import Document
    from docprompt.schema.pipeline import DocumentNode


service_account_file_read_lock = Lock()


orientation_rotation_mapping = {
    0: 0,
    1: 0,
    2: 90,
    3: 180,
    4: -90,
}

service_account_file_read_lock = Lock()

# This will wait up to ~8 minutes before giving up, which covers almost all high-contention cases
# TODO: Scope this to only retry on 429 and 5xx
default_retry_decorator = retry(
    wait=wait_exponential(multiplier=1, max=60), stop=stop_after_attempt(10)
)


def bounding_poly_from_layout(
    layout: Union["documentai.Document.Page.Layout", "documentai.Document.Page.Token"],
):
    return BoundingPoly(
        normalized_vertices=[
            Point(x=round(vertex.x, 5), y=round(vertex.y, 5))
            for vertex in layout.bounding_poly.normalized_vertices
        ]
    )


def bounding_box_from_layout(
    layout: Union["documentai.Document.Page.Layout", "documentai.Document.Page.Token"],
):
    sorted_vertices = sorted(
        layout.bounding_poly.normalized_vertices, key=lambda x: (x.x, x.y)
    )
    upper_left = sorted_vertices[0]
    lower_right = sorted_vertices[-1]

    return NormBBox(
        x0=round(upper_left.x, 5),
        top=round(upper_left.y, 5),
        x1=round(lower_right.x, 5),
        bottom=round(lower_right.y, 5),
    )


def geometry_from_layout(
    layout: Union["documentai.Document.Page.Layout", "documentai.Document.Page.Token"],
    exclude_bounding_poly: bool = False,
):
    bounding_poly = None if exclude_bounding_poly else bounding_poly_from_layout(layout)

    bounding_box = bounding_box_from_layout(layout)

    return {
        "bounding_poly": bounding_poly,
        "bounding_box": bounding_box,
    }


def text_from_layout(
    layout: Union["documentai.Document.Page.Layout", "documentai.Document.Page.Token"],
    document_text: str,
    offset: int = 0,
) -> str:
    """
    Offset is used to account for the fact that text references
    are relative to the entire document.
    """
    working_text = ""

    for segment in sorted(layout.text_anchor.text_segments, key=lambda x: x.end_index):
        start = getattr(segment, "start_index", 0)
        end = segment.end_index

        working_text += document_text[start - offset : end - offset]

    return working_text


def text_spans_from_layout(
    layout: Union["documentai.Document.Page.Layout", "documentai.Document.Page.Token"],
    level: Literal["page", "document"],
    offset: int = 0,
) -> list[TextSpan]:
    text_spans = []

    for segment in sorted(layout.text_anchor.text_segments, key=lambda x: x.end_index):
        start = getattr(segment, "start_index", 0)
        end = segment.end_index

        text_spans.append(
            TextSpan(
                start_index=start - offset,
                end_index=end - offset,
                level=level,
            )
        )

    return text_spans


def text_blocks_from_page(
    page: "documentai.Document.Page",
    document_text: str,
    type: Literal["line", "block", "token", "paragraph"],
    *,
    exclude_bounding_poly: bool = False,
) -> list[TextBlock]:
    text_blocks = []

    type_mapping: Dict[str, SegmentLevels] = {
        "line": "line",
        "paragraph": "block",
        "block": "block",
        "token": "word",
    }

    orientation_mapping: Dict[int, DirectionChoices] = {
        1: "UP",
        2: "RIGHT",
        3: "DOWN",
        4: "LEFT",
    }

    # Offset is used to account for the fact that text references are relative to the entire document.
    # while we need to compute spans relative to the page.
    offset_low = page.layout.text_anchor.text_segments[0].start_index or 0

    for item in getattr(page, f"{type}s"):
        layout = item.layout
        block_text = text_from_layout(layout, document_text)
        geometry_kwargs = geometry_from_layout(
            layout, exclude_bounding_poly=exclude_bounding_poly
        )
        confidence = layout.confidence
        orientation = orientation_mapping.get(layout.orientation, "UP")

        text_spans = text_spans_from_layout(layout, level="page", offset=offset_low)

        block_type = type_mapping[type]
        text_blocks.append(
            TextBlock(
                text=block_text,
                type=block_type,
                bounding_box=geometry_kwargs["bounding_box"],
                bounding_poly=geometry_kwargs["bounding_poly"],
                metadata=TextBlockMetadata(
                    direction=orientation,
                    confidence=round(confidence, 5),
                ),
                text_spans=text_spans,
            )
        )

    return text_blocks


def process_page(
    document_text: str,
    page,
    doc_page_num: int,
    provider_name: str,
    document_name: str,
    file_hash: str,
    exclude_bounding_poly: bool = False,
) -> OcrPageResult:
    layout = page.layout

    page_text = text_from_layout(layout, document_text)

    word_boxes = text_blocks_from_page(
        page, document_text, "token", exclude_bounding_poly=exclude_bounding_poly
    )
    line_boxes = text_blocks_from_page(
        page, document_text, "line", exclude_bounding_poly=exclude_bounding_poly
    )
    block_boxes = text_blocks_from_page(
        page, document_text, "block", exclude_bounding_poly=exclude_bounding_poly
    )

    return OcrPageResult(
        provider_name=provider_name,
        document_name=document_name,
        file_hash=file_hash,
        page_number=doc_page_num,
        page_text=page_text,
        word_level_blocks=word_boxes,
        line_level_blocks=line_boxes,
        block_level_blocks=block_boxes,
    )


def gcp_documents_to_result_single(
    documents: list["documentai.Document"],
    provider_name: str,
    document_name: str,
    file_hash: str,
    *,
    exclude_bounding_poly: bool = False,
):
    page_offset = 1  # We want pages to be 1-indexed

    results: Dict[int, OcrPageResult] = {}

    for document in tqdm.tqdm(documents):
        for doc_page_num, page in enumerate(document.pages):
            page_result = process_page(
                document.text,
                page,
                page_offset + doc_page_num,
                provider_name,
                document_name,
                file_hash,
                exclude_bounding_poly=exclude_bounding_poly,
            )

            results[page_offset + doc_page_num] = page_result

        page_offset += doc_page_num + 1

    return results


def multi_process_page(args):
    idx, document, provider_name, page_offset, exclude_bounding_poly = args

    results: Dict[int, OcrPageResult] = {}

    for doc_page_num, page in enumerate(document.pages):
        page_result = process_page(
            document.text,
            page,
            page_offset + doc_page_num,
            provider_name,
            exclude_bounding_poly=exclude_bounding_poly,
        )

        results[page_offset + doc_page_num] = page_result

    return idx, results


def gcp_documents_to_result_multi(
    documents: List["documentai.Document"],
    provider_name: str,
    *,
    exclude_bounding_poly: bool = False,
):
    tasks = []
    page_offset = 1  # Pages are 1-indexed

    # Prepare tasks for processing, each with a document index
    for idx, document in enumerate(documents):
        args = (
            idx,
            document,
            provider_name,
            page_offset,
            exclude_bounding_poly,
        )
        tasks.append(args)
        page_offset += len(document.pages)

    document_results = [None] * len(documents)

    worker_count = min(len(documents), min(4, max(multiprocessing.cpu_count(), 1)))

    futures = []

    with tqdm.tqdm(total=len(documents), desc="Processing documents") as pbar:
        with ProcessPoolExecutor(max_workers=worker_count) as executor:
            for task in tasks:
                future = executor.submit(multi_process_page, task)
                future.add_done_callback(lambda x: pbar.update(1))
                futures.append(future)

            for future in as_completed(futures):
                idx, page_results = future.result()
                document_results[idx] = page_results

    results: Dict[int, OcrPageResult] = {}

    for document_result in document_results:
        results.update(document_result)  # type: ignore

    return results


def gcp_documents_to_result(
    documents: List["documentai.Document"],
    provider_name: str,
    document_name: str,
    file_hash: str,
    *,
    mode: Literal["single", "multi"] = "single",
    exclude_bounding_poly: bool = False,
) -> Dict[int, OcrPageResult]:
    if mode == "single" or len(documents) == 1 or multiprocessing.cpu_count() == 1:
        logger.info("Using single process")
        return gcp_documents_to_result_single(
            documents,
            provider_name,
            document_name,
            file_hash,
            exclude_bounding_poly=exclude_bounding_poly,
        )
    elif mode == "multi":
        logger.info("Using multiprocessing")
        return gcp_documents_to_result_multi(
            documents,
            provider_name,
            exclude_bounding_poly=exclude_bounding_poly,
        )
    else:
        raise ValueError("Invalid mode for GCP document processing")


class GoogleOcrProvider(AbstractTaskProvider):
    name = "Google Document AI"
    capabilities = [
        CAPABILITIES.PAGE_TEXT_OCR.value,
        CAPABILITIES.PAGE_LAYOUT_OCR.value,
        CAPABILITIES.PAGE_RASTERIZATION.value,
    ]

    max_bytes_per_request = (
        1024 * 1024 * 20
    )  # 20MB is the max size for a single sync request
    max_page_count = 15

    def __init__(
        self,
        project_id: str,
        processor_id: str,
        *,
        service_account_info: Optional[dict] = None,
        service_account_file: Optional[str] = None,
        location: str = "us",
        max_workers: int = multiprocessing.cpu_count() * 2,
        exclude_bounding_poly: bool = False,
    ):
        if service_account_info is None and service_account_file is None:
            raise ValueError(
                "You must provide either service_account_info or service_account_file"
            )
        if service_account_info is not None and service_account_file is not None:
            raise ValueError(
                "You must provide either service_account_info or service_account_file, not both"
            )

        self.project_id = project_id
        self.processor_id = processor_id
        self.location = location

        self.max_workers = max_workers

        self.service_account_info = service_account_info
        self.service_account_file = service_account_file

        self.exclude_bounding_poly = exclude_bounding_poly

        try:
            from google.cloud import documentai

            self.documentai = documentai
        except ImportError:
            raise ImportError(
                "Please install 'google-cloud-documentai' to use the GoogleCloudVisionTextExtractionProvider"
            )

    @classmethod
    def from_service_account_file(
        cls,
        project_id: str,
        processor_id: str,
        service_account_file: str,
    ):
        return cls(
            project_id,
            processor_id,
            service_account_file=service_account_file,
        )

    def get_documentai_client(self, client_option_kwargs: dict = {}, **kwargs):
        from google.api_core.client_options import ClientOptions

        opts = ClientOptions(
            **{
                "api_endpoint": "us-documentai.googleapis.com",
                **client_option_kwargs,
            }
        )

        base_service_client_kwargs = {
            **kwargs,
            "client_options": opts,
        }

        if self.service_account_info is not None:
            return self.documentai.DocumentProcessorServiceClient.from_service_account_info(
                info=self.service_account_info,
                **base_service_client_kwargs,
            )
        elif self.service_account_file is not None:
            with service_account_file_read_lock:
                return self.documentai.DocumentProcessorServiceClient.from_service_account_file(
                    filename=self.service_account_file,
                    **base_service_client_kwargs,
                )
        else:
            raise ValueError("Missing account info and service file path.")

    def _process_document_sync(self, document: Document):
        """
        Split the document into chunks of 15 pages or less, and process each chunk
        synchronously.
        """
        client = self.get_documentai_client()
        processor_name = client.processor_path(
            project=self.project_id,
            location=self.location,
            processor=self.processor_id,
        )

        documents: List["documentai.Document"] = []

        file_bytes = document.get_bytes()

        @default_retry_decorator
        def process_byte_chunk(split_bytes: bytes) -> "documentai.Document":
            raw_document = self.documentai.RawDocument(
                content=split_bytes,
                mime_type="application/pdf",
            )

            field_mask = (
                "text,pages.layout,pages.words,pages.lines,pages.tokens,pages.blocks"
            )

            request = self.documentai.ProcessRequest(
                name=processor_name, raw_document=raw_document, field_mask=field_mask
            )

            result = client.process_document(request=request)

            return result.document

        with tqdm.tqdm(
            total=len(file_bytes), unit="B", unit_scale=True, desc="Processing document"
        ) as pbar:
            for split_bytes in pdf_split_iter_with_max_bytes(
                file_bytes,
                max_page_count=self.max_page_count,
                max_bytes=self.max_bytes_per_request,
            ):
                document = process_byte_chunk(split_bytes)

                documents.append(document)

                pbar.update(len(split_bytes))

        return gcp_documents_to_result(
            documents,
            self.name,
            document_name=document.name,
            file_hash=document.document_hash,
            exclude_bounding_poly=self.exclude_bounding_poly,
        )

    def _process_document_concurrent(
        self,
        document: Document,
        start: int | None = None,
        stop: int | None = None,
        include_raster: bool = False,
    ):
        # Process page chunks concurrently
        client = self.get_documentai_client()
        processor_name = client.processor_path(
            project=self.project_id,
            location=self.location,
            processor=self.processor_id,
        )

        logger.info("Splitting document into chunks...")
        document_byte_splits = list(
            pdf_split_iter_with_max_bytes(
                document.get_bytes(),
                max_page_count=self.max_page_count,
                max_bytes=self.max_bytes_per_request,
            )
        )

        max_workers = min(len(document_byte_splits), self.max_workers)

        @default_retry_decorator
        def process_byte_chunk(split_bytes: bytes):
            raw_document = self.documentai.RawDocument(
                content=split_bytes,
                mime_type="application/pdf",
            )

            field_mask = (
                "text,pages.layout,pages.words,pages.lines,pages.tokens,pages.blocks"
            )

            request = self.documentai.ProcessRequest(
                name=processor_name, raw_document=raw_document, field_mask=field_mask
            )

            result = client.process_document(request=request)

            document = result.document

            return document

        logger.info(f"Processing {len(document_byte_splits)} chunks...")
        with tqdm.tqdm(
            total=len(document_byte_splits), desc="Processing document"
        ) as pbar:
            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                future_to_index = {
                    executor.submit(process_byte_chunk, split): index
                    for index, split in enumerate(document_byte_splits)
                }

                documents: List["documentai.Document"] = [None] * len(  # type: ignore
                    document_byte_splits
                )

                for future in as_completed(future_to_index):
                    index = future_to_index[future]
                    documents[index] = future.result()
                    pbar.update(1)

        logger.info("Recombining")
        return gcp_documents_to_result(
            documents,
            self.name,
            document_name=document.name,
            file_hash=document.document_hash,
            exclude_bounding_poly=self.exclude_bounding_poly,
        )

    def process_document_pages(
        self,
        document: Document,
        start: int | None = None,
        stop: int | None = None,
        **kwargs,
    ) -> Dict[int, OcrPageResult]:
        return self._process_document_concurrent(document, start=start, stop=stop)

    def contribute_to_document_node(
        self, document_node: "DocumentNode", results: Dict[int, OcrPageResult]
    ) -> None:
        for page_number, result in results.items():
            document_node.page_nodes[page_number - 1].ocr_results.results[self.name] = (
                result
            )



from os import PathLike
from pathlib import Path
from subprocess import PIPE, CompletedProcess, run
from typing import Literal, Union

GS = "gs"


class GhostscriptError(Exception):
    def __init__(self, message: str, process: CompletedProcess) -> None:
        self.process = process
        super().__init__(message)


def compress_pdf(
    fp: Union[PathLike, str],  # Ghostscript insists on a file instead of bytes
    output_path: str,
    *,
    compression: Literal["jpeg", "lossless"] = "jpeg",
):
    compression_args = []
    if compression == "jpeg":
        compression_args = [
            "-dAutoFilterColorImages=false",
            "-dColorImageFilter=/DCTEncode",
            "-dAutoFilterGrayImages=false",
            "-dGrayImageFilter=/DCTEncode",
        ]
    elif compression == "lossless":
        compression_args = [
            "-dAutoFilterColorImages=false",
            "-dColorImageFilter=/FlateEncode",
            "-dAutoFilterGrayImages=false",
            "-dGrayImageFilter=/FlateEncode",
        ]
    else:
        compression_args = [
            "-dAutoFilterColorImages=true",
            "-dAutoFilterGrayImages=true",
        ]

    args_gs = (
        [
            GS,
            "-q",
            "-dBATCH",
            "-dNOPAUSE",
            "-dSAFER",
            "-dCompatibilityLevel=1.5",
            "-sDEVICE=pdfwrite",
            "-dAutoRotatePages=/None",
            "-sColorConversionStrategy=LeaveColorUnchanged",
        ]
        + compression_args
        + [
            "-dJPEGQ=95",
            "-dPDFA=2",
            "-dPDFACompatibilityPolicy=1",
            "-sOutputFile=" + output_path,
            str(fp),
        ]
    )

    result = run(args_gs, stdout=PIPE, stderr=PIPE, check=False)

    if result.returncode != 0:
        raise GhostscriptError("Ghostscript failed to compress the document", result)

    return result


def compress_pdf_to_bytes(
    fp: Union[PathLike, str], *, compression: Literal["jpeg", "lossless"] = "jpeg"
) -> bytes:
    result = compress_pdf(fp, output_path="%stdout", compression=compression)

    return result.stdout


def compress_pdf_to_path(
    fp: Union[PathLike, str],
    output_path: PathLike,
    *,
    compression: Literal["jpeg", "lossless"] = "jpeg",
) -> Path:
    compress_pdf(fp, output_path=str(output_path), compression=compression)

    return Path(output_path)



"""Top-level package for Docprompt."""

__author__ = """Frankie Colson"""
__email__ = "frank@pageleaf.io"
__version__ = "0.2.1"

from docprompt.schema.document import Document, PdfDocument  # noqa
from docprompt.schema.layout import NormBBox, TextBlock  # noqa
from docprompt.schema.pipeline import DocumentCollection, DocumentNode, PageNode  # noqa
from docprompt.utils import load_document  # noqa

Document.model_rebuild()


__all__ = [
    "Document",
    "PdfDocument",
    "DocumentCollection",
    "DocumentNode",
    "NormBBox",
    "PageNode",
    "TextBlock",
    "load_document",
]



from typing import Optional
from pydantic import Field
from docprompt.schema.layout import TextBlock
from docprompt.tasks.base import BasePageResult


class OcrPageResult(BasePageResult):
    page_text: str = Field(description="The text for the entire page in reading order")

    word_level_blocks: list[TextBlock] = Field(
        default_factory=list,
        description="The provider-sourced words for the page",
        repr=False,
    )
    line_level_blocks: list[TextBlock] = Field(
        default_factory=list,
        description="The provider-sourced lines for the page",
        repr=False,
    )
    block_level_blocks: list[TextBlock] = Field(
        default_factory=list,
        description="The provider-sourced blocks for the page",
        repr=False,
    )

    raster_image: Optional[bytes] = Field(
        default=None,
        description="The rasterized image of the page used in OCR",
        repr=False,
    )

    @property
    def words(self):
        return self.word_level_blocks

    @property
    def lines(self):
        return self.line_level_blocks

    @property
    def blocks(self):
        return self.block_level_blocks



from .util import (
    get_page_count,
    is_pdf,
    load_document,
    load_document_from_url,
    load_documents_from_urls,
)

from .date_extraction import extract_dates_from_text


__all__ = [
    "get_page_count",
    "is_pdf",
    "load_document",
    "load_document_from_url",
    "load_documents_from_urls",
    "extract_dates_from_text",
]



from typing import Any, Dict, Generic, Literal, TypeVar, TYPE_CHECKING, Union
from pydantic import BaseModel, Field

from docprompt.schema.document import Document
from typing import Optional
from enum import Enum
from datetime import datetime
import importlib


if TYPE_CHECKING:
    from docprompt.schema.pipeline import DocumentNode
    from langchain.schema import SystemMessage, HumanMessage


class CAPABILITIES(Enum):
    """
    Represents a capability that a provider can fulfill
    """

    PAGE_RASTERIZATION = "page-rasterization"
    PAGE_LAYOUT_OCR = "page-layout-ocr"
    PAGE_TEXT_OCR = "page-text-ocr"
    PAGE_CLASSIFICATION = "page-classification"
    PAGE_SEGMENTATION = "page-segmentation"
    PAGE_VQA = "page-vqa"
    PAGE_TABLE_IDENTIFICATION = "page-table-identification"
    PAGE_TABLE_EXTRACTION = "page-table-extraction"


class BaseResult(BaseModel):
    provider_name: str = Field(
        description="The name of the provider which produced the result"
    )
    when: datetime = Field(
        default_factory=datetime.now, description="The time the result was produced"
    )


class BaseDocumentResult(BaseResult):
    document_name: str = Field(description="The name of the document")
    file_hash: str = Field(description="The hash of the document")


class BasePageResult(BaseDocumentResult):
    page_number: int = Field(description="The page number")


PageTaskResult = TypeVar("PageTaskResult", bound=BasePageResult)
DocumentTaskResult = TypeVar("DocumentTaskResult", bound=BaseDocumentResult)
PageOrDocumentTaskResult = TypeVar("PageOrDocumentTaskResult", bound=BaseResult)


class ResultContainer(BaseModel, Generic[PageOrDocumentTaskResult]):
    """
    Represents a container for results of a task
    """

    results: dict[str, PageOrDocumentTaskResult] = Field(
        description="The results of the task, keyed by provider", default_factory=dict
    )

    @property
    def result(self):
        return next(iter(self.results.values()), None)


class AbstractTaskProvider(Generic[PageTaskResult]):
    """
    A task provider performs a specific, repeatable task on a document or its pages
    """

    name: str
    capabilities: list[str]

    def process_document_pages(
        self,
        document: Document,
        start: Optional[int] = None,
        stop: Optional[int] = None,
        **kwargs,
    ) -> dict[int, PageTaskResult]:
        raise NotImplementedError

    def contribute_to_document_node(
        self,
        document_node: "DocumentNode",
        results: dict[int, PageTaskResult],
    ) -> None:
        """
        Adds the results of this task to the document node and/or its page nodes
        """
        pass

    def process_document_node(
        self,
        document_node: "DocumentNode",
        start: Optional[int] = None,
        stop: Optional[int] = None,
        contribute_to_document: bool = True,
        **kwargs,
    ) -> dict[int, PageTaskResult]:
        results = self.process_document_pages(
            document_node.document, start=start, stop=stop, **kwargs
        )

        if contribute_to_document:
            self.contribute_to_document_node(document_node, results)

        return results


def attempt_import(name: str):
    """
    Attempts to import a module or class by name
    """
    package, obj = name.rsplit(".", 1)

    try:
        module = importlib.import_module(package)
    except ImportError:
        return None

    return getattr(module, obj, None)


SupportedModels = Literal["openai", "openai_async", "langchain"]


def validate_language_model(model: Any):
    langchain_chat_klass = attempt_import(
        "langchain_core.language_models.chat_models.BaseChatModel"
    )

    if langchain_chat_klass and isinstance(model, langchain_chat_klass):
        return "langchain"

    openai_klass = attempt_import("openai.OpenAI")

    if openai_klass and isinstance(model, openai_klass):
        return "openai"

    openai_async_klass = attempt_import("openai.OpenAIAsync")

    if openai_async_klass and isinstance(model, openai_async_klass):
        return "openai_async"

    raise ValueError(
        f"Model must be one of langchain_core.language_models.chat_models.BaseChatModel or openai.OpenAI. Got {type(model)}"
    )


SystemMessageLike = Union["SystemMessage", Dict[str, str], str]
HumanMessageLike = Union["HumanMessage", Dict[str, Union[str, Dict[str, str]]], str]


class AbstractLanguageModelTaskProvider(AbstractTaskProvider):
    """
    Provides additional methods for language model specific tasks
    """

    def __init__(self, language_model: Any, *, model_name: Optional[str] = None):
        self.language_model = language_model
        self.model_type = validate_language_model(language_model)

        self.model_name = model_name

        self._validate_kwargs()

    def _validate_kwargs(self):
        """
        Validates the kwargs for the language model
        """
        if self.model_type == "openai" and self.model_name is None:
            raise ValueError("model_name must be provided for OpenAI language models")



from docprompt.schema.pipeline import DocumentNode
from docprompt.tasks.base import AbstractLanguageModelTaskProvider, CAPABILITIES
from .result import TableExtractionPageResult


class BaseTableExtractionProvider(AbstractLanguageModelTaskProvider):
    capabilities = [
        CAPABILITIES.PAGE_TABLE_IDENTIFICATION.value,
        CAPABILITIES.PAGE_TABLE_EXTRACTION.value,
    ]

    def contribute_to_document_node(
        self, document_node: DocumentNode, results: dict[int, TableExtractionPageResult]
    ) -> None:
        for page_number, page_result in results.items():
            document_node.page_nodes[page_number - 1].table_extraction_results.results[
                self.name
            ] = page_result



import tempfile
from typing import Literal

from docprompt._exec.ghostscript import compress_pdf_to_bytes


def compress_pdf_bytes(file_bytes: bytes, *, compression: Literal["jpeg", "lossless"] = "jpeg") -> bytes:
    with tempfile.NamedTemporaryFile(suffix=".pdf") as temp_file:
        temp_file.write(file_bytes)
        temp_file.flush()

        return compress_pdf_to_bytes(temp_file.name, compression=compression)



from typing import Literal, Optional

from pydantic import BaseModel, Field

SegmentLevels = Literal["word", "line", "block"]
TextblockSource = Literal["ocr", "derived"]
DirectionChoices = Literal["UP", "DOWN", "LEFT", "RIGHT"]


class TextSpan(BaseModel):
    start_index: int
    end_index: int
    level: Literal["page", "document"] = Field(
        default="page", description="The level of the span"
    )


class NormBBox(BaseModel):
    """
    Represents a normalized bounding box with each value in the range [0, 1]

    Where x1 > x0 and bottom > top
    """

    x0: float
    top: float
    x1: float
    bottom: float

    class Config:
        json_encoders = {float: lambda v: round(v, 5)}  # 1/10,000 increments is plenty

    def as_tuple(self):
        return (self.x0, self.top, self.x1, self.bottom)

    def __getitem__(self, index):
        # Lots of if statements to prevent new allocations
        if index > 3:
            raise IndexError("Index out of range")

        if index == 0:
            return self.x0
        elif index == 1:
            return self.top
        elif index == 2:
            return self.x1
        elif index == 3:
            return self.bottom

    def __eq__(self, other):
        if not isinstance(other, NormBBox):
            return False

        return self.as_tuple() == other.as_tuple()

    def __hash__(self):
        return hash(self.as_tuple())

    def __and__(self, other):
        if not isinstance(other, NormBBox):
            raise TypeError("Can only compute intersection with NormBBox")
        # Compute the intersection of two bounding boxes
        new_x0 = max(self.x0, other.x0)
        new_top = max(self.top, other.top)
        new_x1 = min(self.x1, other.x1)
        new_bottom = min(self.bottom, other.bottom)

        # Check if there is an actual intersection and if the resulting bounding box is valid
        if new_x0 <= new_x1 and new_top <= new_bottom:
            return NormBBox(x0=new_x0, top=new_top, x1=new_x1, bottom=new_bottom)
        else:
            # Return an empty or non-existent bounding box representation
            return None

    def intersection_over_union(self, other):
        if not isinstance(other, NormBBox):
            raise TypeError("Can only compute IOU with NormBBox")

        # Compute the intersection
        intersection_bbox = self & other

        if intersection_bbox:
            intersection_area = intersection_bbox.area
            union_area = self.area + other.area - intersection_area
            return intersection_area / union_area

        return 0  # No intersection

    def __add__(self, other):
        if not isinstance(other, NormBBox):
            raise TypeError("Can only add NormBBox to NormBBox")

        return NormBBox(
            x0=min(self.x0, other.x0),
            top=min(self.top, other.top),
            x1=max(self.x1, other.x1),
            bottom=max(self.bottom, other.bottom),
        )

    @classmethod
    def combine(cls, *bboxes: "NormBBox"):
        """
        Combines multiple bounding boxes into a single bounding box
        """
        if len(bboxes) == 0:
            raise ValueError("Must provide at least one bounding box")

        if len(bboxes) == 1:
            return bboxes[0]

        working_bbox = bboxes[0]
        for bbox in bboxes[1:]:
            working_bbox = working_bbox + bbox

        return working_bbox

    @classmethod
    def from_bounding_poly(cls, bounding_poly: "BoundingPoly"):
        """
        Returns a NormBBox from a BoundingPoly
        """
        if len(bounding_poly.normalized_vertices) != 4:
            raise ValueError(
                "BoundingPoly must have 4 vertices for NormBBox conversion"
            )

        (
            top_left,
            top_right,
            bottom_right,
            bottom_left,
        ) = bounding_poly.normalized_vertices

        return cls(
            x0=top_left.x,
            top=top_left.y,
            x1=bottom_right.x,
            bottom=bottom_right.y,
        )

    @property
    def width(self):
        return self.x1 - self.x0

    @property
    def height(self):
        return self.bottom - self.top

    @property
    def area(self):
        return self.width * self.height

    @property
    def centroid(self):
        return (self.x0 + self.x1) / 2, (self.top + self.bottom) / 2

    @property
    def y_center(self):
        return (self.top + self.bottom) / 2

    @property
    def x_center(self):
        return (self.x0 + self.x1) / 2


class Point(BaseModel):
    """
    Represents a normalized bounding box with each value in the range [0, 1]
    """

    class Config:
        json_encoders = {float: lambda v: round(v, 5)}  # 1/10,000 increments is plenty

    x: float
    y: float


class BoundingPoly(BaseModel):
    """
    Represents a normalized bounding poly with each value in the range [0, 1]

    Used for higher order shapes like polygons on a page
    """

    normalized_vertices: list[Point]

    def __getitem__(self, index):
        return self.normalized_vertices[index]


class TextBlockMetadata(BaseModel):
    direction: Optional[DirectionChoices] = None
    confidence: Optional[float] = None
    layout_category: Optional[str] = Field(
        default=None, description="The category of the text block"
    )


class TextBlock(BaseModel):
    """
    Represents a single block of text, with its bounding box.
    The bounding box is a tuple of (x0, top, x1, bottom) and
    is normalized to the page size.
    """

    class Config:
        json_encoders = {float: lambda v: round(v, 5)}  # 1/10,000 increments is plenty

    text: str
    type: SegmentLevels
    source: TextblockSource = Field(
        default="derived", description="The source of the text block"
    )

    # Layout information
    bounding_box: NormBBox = Field(default=None, repr=False)
    bounding_poly: Optional[BoundingPoly] = Field(default=None, repr=False)
    text_spans: Optional[list[TextSpan]] = Field(default=None, repr=False)

    metadata: Optional[TextBlockMetadata] = Field(default_factory=TextBlockMetadata)

    def __getitem__(self, index):
        return getattr(self, index)

    @property
    def confidence(self):
        return self.metadata.confidence

    @property
    def direction(self):
        return self.metadata.direction



from docprompt.tasks.base import BasePageResult
from pydantic import BaseModel, Field
from typing import Optional
from docprompt.schema.layout import NormBBox


class TableHeader(BaseModel):
    text: str
    bbox: Optional[NormBBox] = None


class TableCell(BaseModel):
    text: str
    bbox: Optional[NormBBox] = None


class TableRow(BaseModel):
    cells: list[TableCell] = Field(default_factory=list)
    bbox: Optional[NormBBox] = None


class ExtractedTable(BaseModel):
    title: Optional[str] = None
    bbox: Optional[NormBBox] = None

    headers: list[TableHeader] = Field(default_factory=list)
    rows: list[TableRow] = Field(default_factory=list)


class TableExtractionPageResult(BasePageResult):
    tables: list[ExtractedTable] = Field(default_factory=list)



import io
import logging
import tempfile
from typing import Iterator, Optional

import pypdfium2 as pdfium

from docprompt._exec.ghostscript import compress_pdf_to_bytes
from docprompt.utils import get_page_count

logger = logging.getLogger(__name__)


class UnsupportedDocumentType(ValueError):
    pass


def split_pdf_to_bytes(
    file_bytes: bytes,
    *,
    start_page: Optional[int] = None,
    stop_page: Optional[int] = None,
):
    """
    Splits a PDF into a list of bytes.
    """
    if start_page is None:
        start_page = 0
    if stop_page is None:
        stop_page = get_page_count(file_bytes)

    if stop_page <= start_page:
        raise ValueError("stop_page must be greater than start_page")

    # Load the PDF from bytes
    src_pdf = pdfium.PdfDocument(io.BytesIO(file_bytes))

    # Create a new PDF for the current batch
    dst_pdf = pdfium.PdfDocument.new()

    # Append pages to the batch
    dst_pdf.import_pages(src_pdf, list(range(start_page, stop_page)))

    # Save the batch PDF to a bytes buffer
    pdf_bytes_buffer = io.BytesIO()
    dst_pdf.save(pdf_bytes_buffer)
    pdf_bytes_buffer.seek(0)  # Reset buffer pointer to the beginning

    # Yield the bytes of the batch PDF
    return pdf_bytes_buffer.getvalue()


def pdf_split_iter_fast(file_bytes: bytes, max_page_count: int) -> Iterator[bytes]:
    """
    Splits a PDF into batches of pages up to `max_page_count` pages quickly.
    """
    # Load the PDF from bytes
    src_pdf = pdfium.PdfDocument(io.BytesIO(file_bytes))

    current_page = 0
    total_pages = len(src_pdf)

    while current_page < total_pages:
        # Determine the last page for the current batch
        last_page = min(current_page + max_page_count, total_pages)

        dst_pdf = pdfium.PdfDocument.new()

        # Append pages to the batch
        dst_pdf.import_pages(src_pdf, list(range(current_page, last_page)))

        # Save the batch PDF to a bytes buffer
        pdf_bytes_buffer = io.BytesIO()
        dst_pdf.save(pdf_bytes_buffer)
        pdf_bytes_buffer.seek(0)  # Reset buffer pointer to the beginning

        # Yield the bytes of the batch PDF
        yield pdf_bytes_buffer.getvalue()

        # Update the current page for the next batch
        current_page += max_page_count


def pdf_split_iter_with_max_bytes(
    file_bytes: bytes, max_page_count: int, max_bytes: int
) -> Iterator[bytes]:
    """
    Splits a PDF into batches of pages up to `max_page_count` pages and `max_bytes` bytes.
    """
    for batch_bytes in pdf_split_iter_fast(file_bytes, max_page_count):
        if len(batch_bytes) <= max_bytes:
            yield batch_bytes
        else:
            # If batch size is greater than max_bytes, reduce the number of pages
            pages_in_batch = max_page_count
            while len(batch_bytes) > max_bytes and pages_in_batch > 1:
                pages_in_batch -= 1
                batch_bytes = next(pdf_split_iter_fast(file_bytes, pages_in_batch))

            if len(batch_bytes) > max_bytes and pages_in_batch == 1:
                # If a single page is still too large, compress it
                with tempfile.NamedTemporaryFile(suffix=".pdf") as f:
                    f.write(batch_bytes)
                    f.flush()
                    compressed_bytes = compress_pdf_to_bytes(f.name)
                yield compressed_bytes
            else:
                yield batch_bytes



from pydantic import BaseModel, Field, PositiveInt, computed_field
from typing import Optional, List, Literal
from docprompt.schema.layout import TextBlock


class PageTextLocation(BaseModel):
    """
    Specifies the location of a piece of text in a page
    """

    source_blocks: List[TextBlock] = Field(
        description="The source text blocks", repr=False
    )
    text: str  # Sometimes the source text is less than the textblock's text.
    score: float
    granularity: Literal["word", "line", "block"] = "block"

    merged_source_block: Optional[TextBlock] = Field(default=None)


class ProvenanceSource(BaseModel):
    """
    Bundled with some data, specifies exactly where a piece of verbatim text came from
    in a document.
    """

    document_name: str
    page_number: PositiveInt
    text_location: PageTextLocation

    @computed_field  # type: ignore
    @property
    def source_block(self) -> Optional[TextBlock]:
        if self.text_location.merged_source_block:
            return self.text_location.merged_source_block
        if self.text_location.source_blocks:
            return self.text_location.source_blocks[0]

        return None

    @property
    def text(self) -> str:
        if self.text_location:
            return "\n".join([block.text for block in self.text_location.source_blocks])

        return ""



from docprompt.utils.date_extraction import extract_dates_from_text
from datetime import date


STRING_A = """
There was a meeting on 2021-01-01 and another on 2021-01-02.

Meanwhile on September 1, 2021, there was a third meeting.

The final meeting was on 4/5/2021.
"""


def test_date_extraction():
    dates = extract_dates_from_text(STRING_A)

    assert len(dates) == 5

    dates.sort(key=lambda x: x[0])

    assert dates[0][0] == date(2021, 1, 1)
    assert dates[0][1] == "2021-01-01"

    assert dates[1][0] == date(2021, 1, 2)
    assert dates[1][1] == "2021-01-02"

    assert dates[2][0] == date(2021, 4, 5)
    assert dates[2][1] == "4/5/2021"

    assert dates[3][0] == date(2021, 5, 4)
    assert dates[3][1] == "4/5/2021"

    assert dates[4][0] == date(2021, 9, 1)
    assert dates[4][1] == "September 1, 2021"



from pydantic import BaseModel, Field, PositiveInt, TypeAdapter
from pathlib import Path
from typing import Dict, Optional

from docprompt.tasks.ocr.result import OcrPageResult


FIXTURE_PATH = Path(__file__).parent / "fixtures"


OCR_ADAPTER = TypeAdapter(Dict[int, OcrPageResult])


class PdfFixture(BaseModel):
    name: str = Field(description="The name of the fixture")
    page_count: PositiveInt = Field(description="The number of pages in the fixture")
    file_hash: str = Field(description="The expected hash of the fixture")
    ocr_name: Optional[str] = Field(
        description="The path to the OCR results for the fixture", default=None
    )

    def get_full_path(self):
        return FIXTURE_PATH / self.name

    def get_bytes(self):
        return self.get_full_path().read_bytes()

    def get_ocr_results(self):
        if not self.ocr_name:
            return None

        ocr_path = FIXTURE_PATH / self.ocr_name

        return OCR_ADAPTER.validate_json(ocr_path.read_text())



from dataclasses import dataclass, field
from typing import Any, Dict, List, Literal, TYPE_CHECKING, Optional

from docprompt.schema.layout import NormBBox, TextBlock
from docprompt.tasks.ocr.result import OcrPageResult
from .source import PageTextLocation, ProvenanceSource
from .util import (
    create_tantivy_document_wise_block_index,
    preprocess_query_text,
    insert_generator,
    construct_valid_rtree_tuple,
    refine_block_to_word_level,
)

try:
    import tantivy
    from rtree.index import Index as RTreeIndex
except ImportError:
    raise ImportError(
        "Could not import tantivy and/or rtree. Install with `docprompt[search]`"
    )


if TYPE_CHECKING:
    from docprompt.schema.pipeline import DocumentNode


SearchBestModes = Literal["shortest_text", "longest_text", "highest_score"]
BlockGranularity = Literal["word", "line", "block"]
DocumentProvenanceGeoMap = Dict[int, Dict[BlockGranularity, RTreeIndex]]


@dataclass
class DocumentProvenanceLocator:
    document_name: str
    search_index: Any
    block_mapping: Dict[int, OcrPageResult] = field(repr=False)
    geo_index: DocumentProvenanceGeoMap = field(repr=False)

    @classmethod
    def from_document_node(cls, document_node: "DocumentNode"):
        index = create_tantivy_document_wise_block_index()
        block_mapping_dict = {}
        geo_index_dict: DocumentProvenanceGeoMap = {}

        writer = index.writer()

        for page_node in document_node.page_nodes:
            if (
                not page_node.ocr_results.result
                or not page_node.ocr_results.result.block_level_blocks
            ):
                continue

            ocr_result = page_node.ocr_results.result

            for idx, text_block in enumerate(ocr_result.block_level_blocks):
                writer.add_document(
                    tantivy.Document(
                        page_number=page_node.page_number,
                        block_type=text_block.type,
                        block_page_idx=idx,
                        content=text_block.text,
                    )
                )

            for granularity in ["word", "line", "block"]:
                text_blocks = getattr(ocr_result, f"{granularity}_level_blocks", [])

                bounding_boxes = [text_block.bounding_box for text_block in text_blocks]

                if bounding_boxes:
                    r_tree = RTreeIndex(
                        insert_generator(bounding_boxes), fill_factor=0.9
                    )
                else:
                    r_tree = RTreeIndex()

                if page_node.page_number not in geo_index_dict:
                    geo_index_dict[page_node.page_number] = {}

                geo_index_dict[page_node.page_number][granularity] = r_tree  # type: ignore

            block_mapping_dict[page_node.page_number] = ocr_result

        writer.commit()
        index.reload()

        return cls(
            document_name=document_node.document.name,
            search_index=index,
            block_mapping=block_mapping_dict,
            geo_index=geo_index_dict,
        )

    def _construct_tantivy_query(
        self, query: str, page_number: Optional[int] = None
    ) -> tantivy.Query:
        query = preprocess_query_text(query)

        if page_number is None:
            return self.search_index.parse_query(f'content:"{query}"')
        else:
            return self.search_index.parse_query(
                f'(page_number:{page_number}) AND content:"{query}"'
            )

    def get_k_nearest_blocks(
        self,
        bbox: NormBBox,
        page_number: int,
        k: int,
        granularity: BlockGranularity = "block",
    ) -> List[TextBlock]:
        """
        Get the k nearest text blocks to a given bounding box
        """
        search_tuple = construct_valid_rtree_tuple(bbox)

        word_level_bbox_indices = list(
            self.geo_index[page_number][granularity].nearest(
                search_tuple, num_results=k
            )
        )

        block_mapping = self.block_mapping[page_number]

        nearest_blocks = [
            getattr(block_mapping, granularity + "s")[idx]
            for idx in word_level_bbox_indices
        ]

        nearest_blocks.sort(key=lambda x: (x.bounding_box.top, x.bounding_box.x0))

        return [x for x in nearest_blocks if x.bounding_box != bbox]

    def get_overlapping_blocks(
        self, bbox: NormBBox, page_number: int, granularity: BlockGranularity = "block"
    ) -> List[TextBlock]:
        """
        Get the text blocks that overlap with a given bounding box
        """
        search_tuple = construct_valid_rtree_tuple(bbox)

        bbox_indices = list(
            self.geo_index[page_number][granularity].intersection(search_tuple)
        )

        block_mapping = self.block_mapping[page_number]

        overlapping_blocks = [
            getattr(block_mapping, f"{granularity}_level_blocks")[idx]
            for idx in bbox_indices
        ]

        overlapping_blocks.sort(key=lambda x: (x.bounding_box.top, x.bounding_box.x0))

        return [x for x in overlapping_blocks if x.bounding_box != bbox]

    def search_raw(self, raw_query: str) -> List[str]:
        """
        Search for a piece of text using a raw query

        Args:
            query: The text to search for
            page_number: The page number to search on
        """
        parsed_query = self.search_index.parse_query(raw_query)

        searcher = self.search_index.searcher()

        search_results = searcher.search(parsed_query, limit=100)

        results = []

        for score, doc_address in search_results.hits:
            doc = searcher.doc(doc_address)

            result_page_number = doc["page_number"][0]
            result_block_page_idx = doc["block_page_idx"][0]
            block_mapping = self.block_mapping[result_page_number]

            source_block: TextBlock = block_mapping.block_level_blocks[
                result_block_page_idx
            ]

            results.append(source_block.text)

        return results

    def refine_query_to_word_level(
        self, query: str, page_number: int, enclosing_block: TextBlock
    ):
        """
        Refine a query to the word level
        """
        search_tuple = construct_valid_rtree_tuple(enclosing_block.bounding_box)

        word_level_bbox_indices = list(
            self.geo_index[page_number]["word"].intersection(search_tuple)
        )
        word_level_blocks_in_original_bbox = [
            self.block_mapping[page_number].word_level_blocks[idx]
            for idx in word_level_bbox_indices
        ]

        refine_result = refine_block_to_word_level(
            source_block=enclosing_block,
            intersecting_word_level_blocks=word_level_blocks_in_original_bbox,
            query=query,
        )

        return refine_result

    def search(
        self,
        query: str,
        page_number: Optional[int] = None,
        refine_to_word: bool = True,
        require_exact_match: bool = True,
    ) -> List[ProvenanceSource]:
        """
        Search for a piece of text in the document and return the source of it

        Args:
            query: The text to search for
            page_number: The page number to search on
            refine_to_word: Whether to refine the search to the word level
            require_exact_match: Whether to require null results if `refine_to_word` is True and no exact match is found
        """
        search_query = self._construct_tantivy_query(query, page_number)

        searcher = self.search_index.searcher()

        search_results = searcher.search(search_query, limit=100)

        results = []

        for score, doc_address in search_results.hits:
            doc = searcher.doc(doc_address)

            result_page_number = doc["page_number"][0]
            result_block_page_idx = doc["block_page_idx"][0]
            block_mapping = self.block_mapping[result_page_number]

            source_block: TextBlock = block_mapping.block_level_blocks[
                result_block_page_idx
            ]

            source_blocks = [source_block]
            principal_block = source_block

            if refine_to_word:
                refine_result = self.refine_query_to_word_level(
                    query=query,
                    page_number=result_page_number,
                    enclosing_block=source_block,
                )

                if refine_result is not None:
                    principal_block, source_blocks = refine_result
                elif require_exact_match:
                    continue

            source = ProvenanceSource(
                document_name=self.document_name,
                page_number=result_page_number,
                text_location=PageTextLocation(
                    source_blocks=source_blocks,
                    text=query,
                    score=score,
                    granularity="block",
                    merged_source_block=principal_block,
                ),
            )
            results.append(source)

        results.sort(key=lambda x: x.page_number)

        return results

    def search_n_best(
        self, query: str, n: int = 3, mode: SearchBestModes = "shortest_text"
    ) -> List[ProvenanceSource]:
        results = self.search(query)

        if not results:
            return []

        if mode == "shortest_text":
            score_func = lambda x: len(x.source_block.text)  # noqa: E731
        elif mode == "longest_text":
            score_func = lambda x: -len(x[0].source_block.text)  # noqa: E731
        elif mode == "highest_score":
            score_func = lambda x: x[1]  # noqa: E731
        else:
            raise ValueError(f"Unknown mode {mode}")

        results.sort(key=score_func)

        return results[:n]



from .fixtures import PDF_FIXTURES

from docprompt import DocumentNode, load_document
from pytest import raises


def test_search():
    document = load_document(PDF_FIXTURES[0].get_full_path())
    document_node = DocumentNode.from_document(document)

    ocr_results = PDF_FIXTURES[0].get_ocr_results()

    with raises(ValueError):
        document_node.refresh_locator()

    for page_num, ocr_results in ocr_results.items():
        document_node.page_nodes[page_num - 1].ocr_results.results[
            ocr_results.provider_name
        ] = ocr_results

    assert document_node._locator is None  # Ensure the locator is not set

    locator = document_node.locator

    result = locator.search("word that doesn't exist")

    assert len(result) == 0

    result_all_pages = locator.search("and")

    assert len(result_all_pages) == 50

    result_page_1 = locator.search("rooted", page_number=1)

    assert len(result_page_1) == 1



from typing import Union

from docprompt.schema.document import PdfDocument
from docprompt.tasks.message import OpenAIMessage, OpenAIComplexContent, OpenAIImageURL
from ..base import BaseTableExtractionProvider
from ..result import (
    ExtractedTable,
    TableRow,
    TableHeader,
    TableCell,
    TableExtractionPageResult,
)
import xml.etree.ElementTree as ET
import re


SYSTEM_PROMPT = """
You are given an image. Identify and extract all tables from the document.

For each table, respond in the following format:

<table>
<title>(value)</title>
<headers>
<header>(value)</header>
...
</headers>
<rows>
<row>
<column>(value)</column>
...
</row>
...
</table>
""".strip()


def _title_from_tree(tree: ET.Element) -> Union[str, None]:
    title = tree.find("title")
    if title is not None:
        return title.text
    return None


def _headers_from_tree(tree: ET.Element) -> list[TableHeader]:
    headers = tree.find("headers")
    if headers is not None:
        return [
            TableHeader(text=header.text or "") for header in headers.findall("header")
        ]
    return []


def _rows_from_tree(tree: ET.Element) -> list[TableRow]:
    rows = tree.find("rows")
    if rows is not None:
        return [
            TableRow(
                cells=[
                    TableCell(text=cell.text or "") for cell in row.findall("column")
                ]
            )
            for row in rows.findall("row")
        ]
    return []


def _find_start_indices(s: str, sub: str) -> list[int]:
    return [m.start() for m in re.finditer(sub, s)]


def _find_end_indices(s: str, sub: str) -> list[int]:
    return [m.end() for m in re.finditer(sub, s)]


def parse_response(response: str) -> list[ExtractedTable]:
    table_start_indices = _find_start_indices(response, "<table>")
    table_end_indices = _find_end_indices(response, "</table>")

    tables: list[ExtractedTable] = []

    for table_start, table_end in zip(table_start_indices, table_end_indices):
        table_str = response[table_start:table_end]

        table_element = ET.fromstring(table_str)

        title = _title_from_tree(table_element)
        headers = _headers_from_tree(table_element)
        rows = _rows_from_tree(table_element)

        tables.append(ExtractedTable(title=title, headers=headers, rows=rows))

    return tables


class ClaudeTableExtractionProvider(BaseTableExtractionProvider):
    def process_document_pages(
        self,
        document: PdfDocument,
        start: int | None = None,
        stop: int | None = None,
        **kwargs,
    ) -> dict[int, TableExtractionPageResult]:
        messages = []
        for page_number in range(start or 1, (stop or len(document)) + 1):
            rastered_page = document.rasterize_page_to_data_uri(page_number)

            messages.append(
                [
                    OpenAIMessage(
                        role="system",
                        content=SYSTEM_PROMPT,
                    ),
                    OpenAIMessage(
                        role="user",
                        content=[
                            OpenAIComplexContent(
                                type="image_url",
                                image_url=OpenAIImageURL(url=rastered_page),
                            )
                        ],
                    ),
                ]
            )

        return {}



from collections import defaultdict
import re
from docprompt.schema.layout import NormBBox, TextBlock
from typing import Any, Iterable, List, Optional
from rapidfuzz import fuzz

try:
    import tantivy
except ImportError:
    raise ImportError("Could not import tantivy. Install with `docprompt[search]`")

try:
    import networkx
except ImportError:
    raise ImportError("Could not import networkx. Install with `docprompt[search]`")


_prefix_regexs = [
    re.compile(r"^\d+\.\s+"),
    re.compile(r"^\d+\.\d+\s+"),
    re.compile(r"^\*+\s+"),
    re.compile(r"^-+\s+"),
]


def preprocess_query_text(text: str) -> str:
    """
    Improve matching ability by applying some preprocessing to the query text.
    """
    for regex in _prefix_regexs:
        text = regex.sub("", text)

    text = text.strip()

    text = text.replace('"', "")

    return text


def fuzzify_token(
    token: str,
    lowercase: bool = True,
    strip_punct: bool = True,
    strip_whitespace: bool = True,
) -> str:
    """
    Convert a token into a form that is suitable for fuzzy matching.
    """
    if strip_whitespace:
        token = token.strip()

    if strip_punct:
        token = token.strip(".")
        token = token.strip(",")

    if lowercase:
        token = token.lower()

    return token


def word_tokenize(text: str) -> List[str]:
    """
    Tokenize a string into words.
    """
    return re.split(r"\s+", text)


def create_tantivy_document_wise_block_index():
    schema_builder = tantivy.SchemaBuilder()

    schema_builder.add_integer_field(
        "page_number", stored=True, indexed=True, fast=True
    )
    schema_builder.add_text_field("block_type", stored=True)
    schema_builder.add_integer_field("block_page_idx", stored=True)
    schema_builder.add_text_field("content", stored=True)

    schema = schema_builder.build()

    index = tantivy.Index(schema=schema)

    return index


def construct_valid_rtree_tuple(bbox: NormBBox):
    # For some reason sometimes the bounding box is invalid (top > bottom, x0 > x1
    # This function is to ensure that the bounding box is valid for the rtree index

    true_top = min(bbox.top, bbox.bottom)
    true_bottom = max(bbox.top, bbox.bottom)

    true_x0 = min(bbox.x0, bbox.x1)
    true_x1 = max(bbox.x0, bbox.x1)

    return (true_x0, true_top, true_x1, true_bottom)


def insert_generator(bboxes: List[NormBBox], data: Optional[Iterable[Any]] = None):
    """
    Make an iterator that yields tuples of (id, bbox, data) for insertion into an RTree index
    which improves performance massively.
    """
    data = data or [None] * len(bboxes)

    for idx, (bbox, data_item) in enumerate(zip(bboxes, data)):
        yield (idx, construct_valid_rtree_tuple(bbox), data_item)


def refine_block_to_word_level(
    source_block: TextBlock,
    intersecting_word_level_blocks: List[TextBlock],
    query: str,
):
    """
    Create a new text block by merging the intersecting word level blocks that
    match the query.

    """
    intersecting_word_level_blocks.sort(
        key=lambda x: (x.bounding_box.top, x.bounding_box.x0)
    )

    tokenized_query = word_tokenize(query)

    if len(tokenized_query) == 1:
        fuzzified = fuzzify_token(tokenized_query[0])
        for word_level_block in intersecting_word_level_blocks:
            if fuzz.ratio(fuzzified, fuzzify_token(word_level_block.text)) > 87.5:
                return word_level_block, [word_level_block]
    else:
        # Populate the block mapping
        token_block_mapping = defaultdict(set)

        first_word = tokenized_query[0]
        last_word = tokenized_query[-1]

        for token in tokenized_query:
            fuzzified_token = fuzzify_token(token)
            for i, word_level_block in enumerate(intersecting_word_level_blocks):
                fuzzified_block_text = fuzzify_token(word_level_block.text)
                if fuzz.ratio(fuzzified_token, fuzzified_block_text) > 87.5:
                    token_block_mapping[token].add(i)

        graph = networkx.DiGraph()
        prev = tokenized_query[0]

        for i in token_block_mapping[prev]:
            graph.add_node(i)

        for token in tokenized_query[1:]:
            for prev_block in token_block_mapping[prev]:
                for block in sorted(token_block_mapping[token]):
                    if block > prev_block:
                        weight = (
                            (block - prev_block) ** 2
                        )  # Square the distance to penalize large jumps, which encourages reading order
                        graph.add_edge(prev_block, block, weight=weight)

            prev = token

        # Get every combination of first and last word
        first_word_blocks = token_block_mapping[first_word]
        last_word_blocks = token_block_mapping[last_word]

        combinations = sorted(
            [(x, y) for x in first_word_blocks for y in last_word_blocks if x < y],
            key=lambda x: abs(x[1] - x[0]),
        )

        for start, end in combinations:
            try:
                path = networkx.shortest_path(graph, start, end, weight="weight")
            except networkx.NetworkXNoPath:
                continue
            except Exception:
                continue

            matching_blocks = [intersecting_word_level_blocks[i] for i in path]

            merged_bbox = NormBBox.combine(
                *[word_level_block.bounding_box for word_level_block in matching_blocks]
            )

            merged_text = ""

            for word_level_block in matching_blocks:
                merged_text += word_level_block.text
                if not word_level_block.text.endswith(" "):
                    merged_text += " "  # Ensure there is a space between words

            return (
                TextBlock(
                    text=merged_text,
                    type="block",
                    bounding_box=merged_bbox,
                    metadata=source_block.metadata,
                ),
                matching_blocks,
            )



from docprompt import load_document
from PIL import Image
import io

from .fixtures import PDF_FIXTURES

from docprompt.utils.splitter import pdf_split_iter_fast


def test_load_document():
    for fixture in PDF_FIXTURES:
        doc = load_document(fixture.get_full_path())
        assert doc.page_count == fixture.page_count
        assert doc.document_hash == fixture.file_hash


def test_rasterize():
    # Fow now just test PIL can open the image
    for fixture in PDF_FIXTURES:
        doc = load_document(fixture.get_full_path())
        img_bytes = doc.rasterize_page(1)
        Image.open(io.BytesIO(img_bytes))


def test_split():
    doc = load_document(PDF_FIXTURES[0].get_full_path())

    new_docs = doc.split(start=2)

    assert len(new_docs) == len(doc) - 2


def test_pdf_split_iter_fast_1_sized():
    doc = load_document(PDF_FIXTURES[0].get_full_path())

    splits = list(pdf_split_iter_fast(doc.file_bytes, 1))

    assert len(splits) == len(doc)



from typing import Generic, Optional, TypeVar, TYPE_CHECKING

from pydantic import BaseModel, Field, PositiveInt, PrivateAttr

from docprompt.tasks.base import ResultContainer
from docprompt.tasks.ocr.result import OcrPageResult

if TYPE_CHECKING:
    from docprompt.provenance.search import DocumentProvenanceLocator

from .document import Document

DocumentCollectionMetadata = TypeVar("DocumentCollectionMetadata", bound=BaseModel)
DocumentNodeMetadata = TypeVar("DocumentNodeMetadata", bound=BaseModel)
PageNodeMetadata = TypeVar("PageNodeMetadata", bound=BaseModel)


class PageNode(BaseModel, Generic[PageNodeMetadata]):
    """
    Represents a single page in a document, with some metadata
    """

    document: "DocumentNode" = Field(exclude=True, repr=False)
    page_number: PositiveInt = Field(description="The page number")
    metadata: Optional[PageNodeMetadata] = Field(
        description="Application-specific metadata for the page", default=None
    )

    ocr_results: ResultContainer[OcrPageResult] = Field(
        default_factory=lambda: ResultContainer(),
        description="The OCR results for the page",
        repr=False,
    )
    table_extraction_results: ResultContainer = Field(
        default_factory=lambda: ResultContainer(),
        description="The table extraction results for the page",
        repr=False,
    )


class DocumentNode(BaseModel, Generic[DocumentNodeMetadata, PageNodeMetadata]):
    """
    Represents a single document, with some metadata
    """

    document: Document
    page_nodes: list[PageNode[PageNodeMetadata]] = Field(
        description="The pages in the document", default_factory=list, repr=False
    )
    metadata: Optional[DocumentNodeMetadata] = Field(
        description="Application-specific metadata for the document", default=None
    )

    _locator: Optional["DocumentProvenanceLocator"] = PrivateAttr(default=None)

    def __len__(self):
        return len(self.page_nodes)

    def __getitem__(self, index):
        return self.page_nodes[index]

    def __iter__(self):
        return iter(self.page_nodes)

    @property
    def locator(self):
        if self._locator is None:
            self.refresh_locator()

        return self._locator

    def refresh_locator(self):
        """
        Refreshes the locator for this document node
        """
        from docprompt.provenance.search import DocumentProvenanceLocator

        if any(not page.ocr_results.result for page in self.page_nodes):
            raise ValueError(
                "Cannot create a locator for a document node with missing OCR results"
            )

        self._locator = DocumentProvenanceLocator.from_document_node(self)

        return self.locator

    @classmethod
    def from_document(
        cls,
        document: Document,
        document_metadata: Optional[DocumentNodeMetadata] = None,
    ):
        document_node: "DocumentNode[DocumentNodeMetadata, PageNodeMetadata]" = (
            DocumentNode(document=document, metadata=document_metadata)
        )

        for page_number in range(1, len(document) + 1):
            document_node.page_nodes.append(
                PageNode(document=document_node, page_number=page_number)
            )

        return document_node

    @property
    def file_hash(self):
        return self.document.document_hash

    @property
    def document_name(self):
        return self.document.name


class DocumentCollection(
    BaseModel,
    Generic[DocumentCollectionMetadata, DocumentNodeMetadata, PageNodeMetadata],
):
    """
    Represents a collection of documents with some common metadata
    """

    document_nodes: list[DocumentNode[DocumentNodeMetadata, PageNodeMetadata]]
    metadata: DocumentCollectionMetadata
